/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Base and common derived exceptions for the jwsc++ library.
 */


#include <jwsc++/config.h>

#include <string>
#include <iostream>

#include <inttypes.h>

#include "jwsc++/base/exception.h"

#ifdef JWSCXX_HAVE_DMALLOC
#include <dmalloc.h>
#endif


using std::string;
using std::ostream;
using std::cerr;
using std::fprintf;
using std::exit;
using std::abort;

using namespace jwscxx::base;


/* --------------------------  Exception  ----------------------------------- */

/**
 * @param  msg   Error message.
 * @param  file  File where the error occurred.
 * @param  line  Line approximately where the error occurred.
 */
Exception::Exception(const char* msg, const char* file, uint32_t line)
    : msg(msg)
{
    this->file = file;
    this->line = line;
}


/**
 * @param  msg   Error message.
 * @param  file  File where the error occurred.
 * @param  line  Line approximately where the error occurred.
 */
Exception::Exception(const string& msg, const char* file, uint32_t line)
    : msg(msg)
{
    this->file = file;
    this->line = line;
}


/** @param  e  Exception to copy into this one. */
Exception::Exception(const Exception& e)
    : msg(e.msg)
{
    file = e.file;
    line = e.line;
}


/**
 * @param  out      Output stream to print to.
 * @param  newline  Whether to print a newline character.
 */
void Exception::print(ostream& out, bool newline) const
{
    out << msg;

    if (newline) out << "\n";
}


/**
 * @param  out      Output stream to print to.
 * @param  newline  Whether to print a newline character.
 */
void Exception::print_details(ostream& out, bool newline) const
{
    if (file && line)
    {
        out << file << " line " << line << ": " << msg;
        if (newline) out << "\n";
    }
    else if (file)
    {
        out << file << ": " << msg;
        if (newline) out << "\n";
    }
    else
    {
        print(out, newline);
    }
}


/**
 * @param  out      Output stream to print to.
 * @param  newline  Whether to print a newline character.
 */
void Exception::print_abort(ostream& out, bool newline) const
{
    print(out, newline);
    abort();
}


/**
 * @param  out      Output stream to print to.
 * @param  newline  Whether to print a newline character.
 */
void Exception::print_details_abort(ostream& out, bool newline) const
{
    print_details(out, newline);
    abort();
}


/**
 * @param  out      Output stream to print to.
 * @param  newline  Whether to print a newline character.
 * @param  status   Exit status.
 */
void Exception::print_exit(ostream& out, bool newline, int status) const
{
    print(out, newline);
    exit(status);
}


/**
 * @param  out      Output stream to print to.
 * @param  newline  Whether to print a newline character.
 * @param  status   Exit status.
 */
void Exception::print_details_exit(ostream& out, bool newline, int status) const
{
    print_details(out, newline);
    exit(status);
}

/* -------------------------------------------------------------------------- */







/* --------------------------  Exception  ----------------------------------- */

/**
 * @param  msg   Error message.
 * @param  file  File where the error occurred.
 * @param  line  Line approximately where the error occurred.
 */
IO_error::IO_error(const char* msg, const char* file, uint32_t line)
    : Exception(msg, file, line)
{
}


/**
 * @param  msg   Error message.
 * @param  file  File where the error occurred.
 * @param  line  Line approximately where the error occurred.
 */
IO_error::IO_error(const string& msg, const char* file, uint32_t line)
    : Exception(msg, file, line)
{
}


/** @param  e  Exception to copy into this one. */
IO_error::IO_error(const IO_error& e)
    : Exception(e)
{
}

/* -------------------------------------------------------------------------- */







/* --------------------------  Exception  ----------------------------------- */

/**
 * @param  msg   Error message.
 * @param  file  File where the error occurred.
 * @param  line  Line approximately where the error occurred.
 */
Arg_error::Arg_error(const char* msg, const char* file, uint32_t line)
    : Exception(msg, file, line)
{
}


/**
 * @param  msg   Error message.
 * @param  file  File where the error occurred.
 * @param  line  Line approximately where the error occurred.
 */
Arg_error::Arg_error(const string& msg, const char* file, uint32_t line)
    : Exception(msg, file, line)
{
}


/** @param  e  Exception to copy into this one. */
Arg_error::Arg_error(const Arg_error& e)
    : Exception(e)
{
}

/* -------------------------------------------------------------------------- */






/* --------------------------  Exception  ----------------------------------- */

/**
 * @param  msg   Error message.
 * @param  file  File where the error occurred.
 * @param  line  Line approximately where the error occurred.
 */
Result_error::Result_error(const char* msg, const char* file, uint32_t line)
    : Exception(msg, file, line)
{
}


/**
 * @param  msg   Error message.
 * @param  file  File where the error occurred.
 * @param  line  Line approximately where the error occurred.
 */
Result_error::Result_error(const string& msg, const char* file, uint32_t line)
    : Exception(msg, file, line)
{
}


/** @param  e  Exception to copy into this one. */
Result_error::Result_error(const Result_error& e)
    : Exception(e)
{
}

/* -------------------------------------------------------------------------- */
