/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Base and common derived exceptions for the jwsc++ library.
 */


#ifndef EXCEPTION_H
#define EXCEPTION_H


#include <jwsc++/config.h>

#include <cstdlib>
#include <cstdio>
#include <string>
#include <iostream>

#include <inttypes.h>


using std::string;
using std::ostream;
using std::cerr;


namespace jwscxx {
namespace base {


/**
 * @class Exception
 *
 * @brief Base class of all exceptions in the jwsc++ library.
 */
class Exception
{
    public:

        /** @brief Constructs an Exception. */
        Exception(const char* msg, const char* file=0, uint32_t line=0);


        /** @brief Constructs an Exception. */
        Exception(const string& msg, const char* file=0, uint32_t line=0);


        /** @brief Constructs an Exception from another. */
        Exception(const Exception& e);


        /** @brief Deletes an Exception. */
        virtual ~Exception() { }


        /** @brief Returns the error message for the Exception. */
        const string& get_msg() const { return msg; }


        /** @brief Returns the file name where the Exception occurred. */
        const char* get_file() const { return file; }


        /** @brief Returns the line number near where the Exception occurred. */
        uint32_t get_line() const { return line; }


        /** @brief Prints the Exception::msg to an ostream. */
        virtual void print(ostream& out=cerr, bool newline=true) const;


        /** @brief Prints all the fields of the Exception to an ostream. */
        virtual void print_details(ostream& out=cerr, bool newline=true) const;


        /** @brief Prints the Exception::msg to an ostream and aborts. */
        virtual void print_abort(ostream& out=cerr, bool newline=true) const;


        /** 
         * @brief Prints all the fields of the Exception to an ostream and
         * aborts.
         */
        virtual void print_details_abort
        (
            ostream& out     = cerr, 
            bool     newline = true
        )
        const;


        /** @brief Prints the Exception::msg to an ostream and exits. */
        virtual void print_exit
        (
            ostream& out     = cerr, 
            bool     newline = true,
            int      status  = EXIT_FAILURE
        )
        const;


        /** 
         * @brief Prints all the fields of the Exception to an ostream and
         * exits.
         */
        virtual void print_details_exit
        (
            ostream& out     = cerr, 
            bool     newline = true,
            int      status  = EXIT_FAILURE
        )
        const;


    protected:

        /** @brief Message associated with the error causing the exception. */
        string msg;

        /** @brief File name where the Error occurred. */
        const char* file;

        /** @brief Line number where the Error occurred. */
        uint32_t line;
};




/**
 * @class IO_error
 *
 * @brief Input/Output error exception.
 */
class IO_error : public Exception
{
    public:

        /** @brief Constructs an Exception caused by an IO error. */
        IO_error(const char* msg, const char* file=0, uint32_t line=0);


        /** @brief Constructs an Exception caused by an IO error. */
        IO_error(const string& msg, const char* file=0, uint32_t line=0);


        /** @brief Constructs an Exception from another. */
        IO_error(const IO_error& e);


        /** @brief Deletes an IO_error. */
        virtual ~IO_error() { }
};




/**
 * @class Arg_error
 *
 * @brief Argument error exception.
 *
 * Thrown when an invalid argument is provided to a function.
 */
class Arg_error : public Exception
{
    public:

        /** 
         * @brief Constructs an Exception caused by an invalid argument to a
         * function.
         */
        Arg_error(const char* msg, const char* file=0, uint32_t line=0);


        /** 
         * @brief Constructs an Exception caused by an invalid argument to a
         * function.
         */
        Arg_error(const string& msg, const char* file=0, uint32_t line=0);


        /** @brief Constructs an Exception from another. */
        Arg_error(const Arg_error& e);


        /** @brief Deletes an Arg_error. */
        virtual ~Arg_error() { }
};




/**
 * @class Result_error
 *
 * @brief Result error exception.
 *
 * Thrown when an invalid result is generated by a function.
 */
class Result_error : public Exception
{
    public:

        /** 
         * @brief Constructs an Exception caused by an invalid result generated
         * by a function.
         */
        Result_error(const char* msg, const char* file=0, uint32_t line=0);


        /** 
         * @brief Constructs an Exception caused by an invalid result generated
         * by a function.
         */
        Result_error(const string& msg, const char* file=0, uint32_t line=0);


        /** @brief Constructs an Exception from another. */
        Result_error(const Result_error& e);


        /** @brief Deletes a Result_error. */
        virtual ~Result_error() { }
};


}}


#endif
