/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Provides program option support via the command-line or an options 
 * file.
 *
 * A program can use this option support by defining a collection of Options
 * and then calling Options::process().
 *
 * The options can be represented by a long or short name. Both can take an 
 * argument by defining an Option_with_arg. If no argument is called for, use 
 * Option_no_arg.
 */


#ifndef OPTION_H
#define OPTION_H


#include <jwsc++/config.h>

#include <iostream>
#include <list>

#include "jwsc++/base/exception.h"


using std::istream;
using std::ostream;
using std::list;


namespace jwscxx {
namespace base {


/**
 * @class Option
 *
 * @brief Base program option.
 *
 * An option is specified on the command-line as follows
 *
 * @code
 *   --option-string
 *   -O
 * @endcode
 *
 * where 
 *
 * @code
 *   Option::l_name = "option-string"
 *   Option::s_name = 'O'
 * @endcode
 *
 * Specifying an option in a file is similar, but only the Option::l_name
 * can be used, for instance 
 *
 * @code
 *   option-string
 * @endcode
 */
class Option
{
    public:

        /** @brief Constructs an Option. */
        Option(char s_name, const char* l_name, const char* desc) 
            throw (Arg_error);


        /** @brief Deletes an Option. */
        virtual ~Option() { }


        /** @brief Processes an Option from program command-line arguments. */
        virtual int process(int argc, char** argv) const 
            throw (Arg_error) = 0;


        /** @brief Prints an Option to an output stream. */
        virtual void print(ostream& out, int width) const throw (Arg_error) = 0;


    protected:

        /** 
         * @brief Returns whether or not @em arg is the short name of this 
         * Option.
         */
        bool is_s_name_option(const char* arg) const;


        /** 
         * @brief Returns whether or not @em arg is the long name of this
         * Option.
         */
        bool is_l_name_option(const char* arg) const;


    protected:

        /** 
         * @brief Short name of the option.
         *
         * Example '-o', where flag='o'. If not used, set to 0 or '\\0'.
         */ 
        char s_name;

        /** 
         * @brief Long name of the option. 
         *
         * Example '--option', where string="option". If not used, set to 0.
         */
        const char* l_name;

        /** 
         * @brief Description of the option.
         *
         * If not used, set to 0.
         */
        const char* desc;
};




/**
 * @class Option_with_arg
 *
 * @brief Option that takes an argument.
 *
 * An option is specified on the command-line as follows
 *
 * @code
 *   --option-string=arg
 *   -O arg
 * @endcode
 *
 * where 
 *
 * @code
 *   Option::l_name = "option-string"
 *   Option::s_name = 'O'
 * @endcode
 *
 * Specifying an option in a file is similar, but only the Option::l_name
 * can be used, for instance 
 *
 * @code
 *   option-string=arg
 * @endcode
 */
class Option_with_arg : public Option
{
    public:

        /** @brief Constructs an Option that takes an argument. */
        Option_with_arg
        (
            char        s_name, 
            const char* l_name, 
            const char* desc,
            void (*func)(const char* arg) throw (Arg_error)
        ) 
        throw (Arg_error);


        /** @brief Processes an Option from program command-line arguments. */
        int process(int argc, char** argv) const throw (Arg_error);


        /** @brief Prints an Option to an output stream. */
        void print(ostream& out, int width) const throw (Arg_error);


    protected:

        /** @brief Gets the argument for the Option from its long name. */
        const char* get_l_name_arg(const char* str) const;


    protected:

        /** 
         * @brief Function to process an Option with an argument.
         *
         * Must be able to handle @em arg == 0.
         */
        void (*func)(const char* arg) throw (Arg_error);
};




/**
 * @class Option_no_arg
 *
 * @brief Option that does not take an argument.
 *
 * An option is specified on the command-line as follows
 *
 * @code
 *   --option-string
 *   -O
 * @endcode
 *
 * where 
 *
 * @code
 *   Option::l_name = "option-string"
 *   Option::s_name = 'O'
 * @endcode
 *
 * Specifying an option in a file is similar, but only the Option::l_name
 * can be used, for instance 
 *
 * @code
 *   option-string
 * @endcode
 */
class Option_no_arg : public Option
{
    public:

        /** @brief Constructs an Option that does not take an argument. */
        Option_no_arg
        (
            char        s_name, 
            const char* l_name, 
            const char* desc,
            void (*func)(void)
        ) 
        throw (Arg_error);


        /** @brief Processes an Option from program command-line arguments. */
        int process(int argc, char** argv) const throw (Arg_error);


        /** @brief Prints an Option to an output stream. */
        void print(ostream& out, int width) const throw (Arg_error);


    protected:

        /** @brief Function to process an Option with an argument. */
        void (*func)(void);
};




/**
 * @class Options
 *
 * @brief Collection of Options for processing.
 */
class Options
{
    public:

        /** @brief Constructs an empty collection of options. */
        Options() : opts() { }


        /** @brief Deletes a collection of options. */
        ~Options();


        /** @brief Adds an option to the collection. */
        void add(Option* opt);


        /** 
         * @brief Processes the options in the collection from program
         * command-line arguments.
         */
        int process(int argc, char** argv) const throw (Arg_error);


        /** @brief Processes the options in the collection from a file. */
        void process(const char* fname) const throw (IO_error, Arg_error);


        /** 
         * @brief Processes the options in the collection from an
         * input stream. 
         */
        void process(istream& in) const throw (Arg_error);


        /** @brief Prints the options in the collection to an output stream. */
        void print(ostream& out=cerr, int width=20) const;


    private:

        /** @brief Returns whether an argument is an option. */
        bool is_an_option(const char* arg) const;

        /** @brief Collection of Options. */
        list<Option *> opts;
};


}}


#endif
