/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Abstract class to read this object from an input stream.
 */


#include <jwsc++/config.h>

#include <iostream>
#include <sstream>
#include <fstream>
#include <cctype>
#include <cstring>

#include <jwsc++/base/exception.h>
#include <jwsc++/base/readable.h>

#ifdef JWSCXX_HAVE_DMALLOC
#include <dmalloc.h>
#endif


using namespace jwscxx::base;


/** 
 * @param  fname  Input file to read from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em fname.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in the file to read from.
 */
void Readable::read(const char* fname) throw (jwscxx::base::IO_error,
        jwscxx::base::Arg_error)
{
    std::ifstream in;
    std::ostringstream ost;

    in.open(fname);
    if (in.fail())
    {
        ost << fname << ": Could not open file";
        throw IO_error(ost.str());
    }

    try
    {
        read(in);
    }
    catch (IO_error e)
    {
        ost << fname << ": " << e.get_msg();
        throw IO_error(ost.str());
    }

    in.close();
    if (in.fail())
    {
        ost << fname << ": Could not close file";
        throw IO_error(ost.str());
    }
}


/**
 * Reads a line off the input stream as a field with format:
 *
 * @code
 *   field name: field value
 * @endcode
 *
 * Where the separator here is a ':'.
 *
 * @param  in          Input stream to read the field line from.
 * @param  field_name  Name of the field to get the value of.
 * @param  field_buf   Buffer to read the line of input into.
 * @param  buf_len     Length of the buffer.
 * @param  separator   Character separating the field name and value. The
 *                     default is ':'.
 *
 * @throw  jwscxx::base::Arg_error  Either the field is not the one named or 
 *                                  it is not formatted properly.
 * @throw  jwscxx::base::IO_error   Could not read a line from the input stream.
 *
 * @return A constant pointer into @em field_buf where the field value is or 0 
 * if there is no value.
 *
 * @note A terminating character '\\0' is always inserted in the buffer, even 
 * if it would truncate the line.
 */
const char* Readable::read_field_value
(
    std::istream& in,
    const char*   field_name, 
    char*         field_buf,
    size_t        buf_len,
    char          separator
) 
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    using std::ostringstream;
    using std::isspace;
    using std::strlen;
    using std::strncmp;

    in.getline(field_buf, buf_len);
    if (in.fail()) 
    {
        throw IO_error("Could not read field line");
    }

    const char* field_ptr = field_buf;

    while (isspace(*field_ptr) && *field_ptr != 0)
    {
        field_ptr++;
    }

    if (strncmp(field_ptr, field_name, strlen(field_name)) != 0)
    {
        ostringstream ost;
        ost << "Field name '" << field_name << "' not found in '" 
            << field_buf << "'";
        throw Arg_error(ost.str());
    }

    while (*field_ptr != separator && *field_ptr != 0)
    {
        field_ptr++;
    }

    if (*field_ptr == 0)
    {
        ostringstream ost;
        ost << "Field '" << field_buf << "' improperly formatted";
        throw Arg_error(ost.str());
    }

    field_ptr++;

    while (isspace(*field_ptr) && *field_ptr != 0)
    {
        field_ptr++;
    }

    if (*field_ptr == 0)
        return 0;

    return field_ptr;
}


/**
 * Reads a line off the input stream as a field with format:
 *
 * @code
 *   field name: field value
 * @endcode
 *
 * Where the separator here is a ':'.
 *
 * @param  in          Input stream to read the field line from.
 * @param  field_name  Name of the field to get the value of.
 * @param  separator   Character separating the field name and value. The
 *                     default is ':'.
 *
 * @throw  jwscxx::base::Arg_error  Either the field is not the one named or 
 *                                  it is not formatted properly.
 * @throw  jwscxx::base::IO_error   Could not read a line from the input stream.
 *
 * @return A constant pointer to the field value or 0 if there is no value.
 *
 * @note The buffer for reading a field value is shared, so each time the
 * function is called, it will be overwritten.
 *
 * @note The maximum line length to read is 255.
 */
const char* Readable::read_field_value
(
    std::istream& in,
    const char*   field_name, 
    char          separator
) 
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    // TODO This is not thread-safe.
    static char field_buf[256] = {0};

    return read_field_value(in, field_name, field_buf, 256, separator);
}
