/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Camera for 3D viewing in OpenGL.
 *
 * The Camera is defined by a perspective reference point (focal point), a view
 * reference point, an up vector, an aperture, and near and far clipping planes.
 */


#ifndef CAMERA_H
#define CAMERA_H


#include <jwsc++/config.h>

#include <iostream>

#if defined JWSCXX_HAVE_OPENGL
#include <GL/gl.h>
#elif defined JWSCXX_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/gl.h>
#endif

#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/image/image.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/base/cloneable.h>
#include <jwsc++/base/readable.h>
#include <jwsc++/base/writeable.h>


namespace jwscxx {
namespace graphics {


/**
 * @class Camera_f
 *
 * @brief Camera for 3D viewing in OpenGL.
 *
 * Uses single precision, floating point values.
 */
class Camera_f : public jwscxx::base::Cloneable, 
                 public jwscxx::base::Writeable, 
                 public jwscxx::base::Readable
{
    public:

        /** @brief Constructs a Camera_f. */
        Camera_f
        (
            float focal_pt_x,
            float focal_pt_y,
            float focal_pt_z,
            float ref_pt_x,
            float ref_pt_y,
            float ref_pt_z,
            float up_x,
            float up_y,
            float up_z,
            float aperture, 
            float clip_near, 
            float clip_far
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a Camera_f. */
        Camera_f
        (
            const jwsc::Vector_f* focal_pt,
            const jwsc::Vector_f* ref_pt,
            const jwsc::Vector_f* up,
            float                 aperture, 
            float                 clip_near, 
            float                 clip_far
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a camera by copying another. */
        Camera_f(const Camera_f& c);


        /** @brief Reads a camera from an input file. */
        Camera_f(const char* fname) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Reads a camera from an input stream. */
        Camera_f(std::istream& in) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Deletes this Camera_f. */
        virtual ~Camera_f();


        /** @brief Copies a camera into this one. */
        virtual Camera_f& operator= (const Camera_f& c);


        /** @brief Clones this camera. */
        virtual Camera_f* clone() const;


        /** @brief Returns a constant pointer to the camera right (x) vector. */
        const jwsc::Vector_f* get_u() const;


        /** @brief Returns a constant pointer to the camera up (y) vector. */
        const jwsc::Vector_f* get_v() const;


        /** @brief Returns a constant pointer to the camera -view (z) vector. */
        const jwsc::Vector_f* get_n() const;


        /** @brief Returns a constant pointer to the camera focal point. */
        const jwsc::Vector_f* get_prp() const;


        /** @brief Returns a constant pointer to the camera reference point. */
        const jwsc::Vector_f* get_vrp() const;


        /** @brief Returns the distance to the near clipping plane. */
        float get_clipping_near() const;


        /** @brief Returns the distance to the far clipping plane. */
        float get_clipping_far() const;


        /** @brief Returns the focal length. */
        float get_focal_length() const;


        /** @brief Returns the aperature. */
        float get_aperture() const;


        /** 
         * @brief Returns half the height of the image plane at the focal
         * length.
         */
        float get_half_focal_height() const;


        /** 
         * @brief Sets the focal length and updates the aperture by changing 
         * the left and right prp.
         */
        virtual void set_f_with_new_aperture_using_prp(float f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length using the same aperture by changing the 
         * left and right prp.
         */
        virtual void set_f_with_same_aperture_using_prp(float f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length and updates the aperture by changing 
         * the vrp.
         */
        virtual void set_f_with_new_aperture_using_vrp(float f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length using the same aperture by changing
         * the vrp.
         */
        virtual void set_f_with_same_aperture_using_vrp(float f) 
            throw (jwscxx::base::Arg_error);


        /** @brief Sets the vertical aperture. */
        virtual void set_aperture(float aperture) 
            throw (jwscxx::base::Arg_error);


        /** @brief Sets the clipping planes. */
        virtual void set_clipping(float clip_near, float clip_far) 
            throw (jwscxx::base::Arg_error);


        /** @brief Rotates this Camera_f around a vector. */
        virtual void rotate(float angle, float x, float y, float z) 
            throw (jwscxx::base::Arg_error);


        /** @brief Rotates this Camera_f around a vector. */
        virtual void rotate(float angle, const jwsc::Vector_f* r) 
            throw (jwscxx::base::Arg_error);


        /** @brief Translates this Camera_f. */
        virtual void translate(float x, float y, float z);


        /** @brief Translates this Camera_f. */
        virtual void translate(const jwsc::Vector_f* t) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the position, direction and focal length of the 
         * camera.
         */
        virtual void look_at
        (
            float focal_pt_x,
            float focal_pt_y,
            float focal_pt_z,
            float ref_pt_x,
            float ref_pt_y,
            float ref_pt_z,
            float up_x,
            float up_y,
            float up_z
        )
        throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the position, direction and focal length of the 
         * camera.
         */
        void look_at
        (
            const jwsc::Vector_f* focal_pt,
            const jwsc::Vector_f* ref_pt,
            const jwsc::Vector_f* up
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Sets the GL model view matrix with this Camera_f. */
        virtual void set_gl_modelview() const;


        /** @brief Sets the GL projection matrix with this Camera_f. */
        virtual void set_gl_projection() const;


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the camera
         * focal plane.
         */
        void get_image_pt_on_focal_plane
        (
            jwsc::Vector_f** pt_out, 
            float      x, 
            float      y
        ); 


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the camera
         * focal plane.
         */
        void get_image_pt_on_focal_plane
        (
            jwsc::Vector_f**      pt_out, 
            const jwsc::Vector_f* pt_in
        )
        throw (jwscxx::base::Arg_error); 


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the near
         * camera clipping plane.
         */
        void get_image_pt_on_near_clipping_plane
        (
            jwsc::Vector_f** pt_out, 
            float            x, 
            float            y
        ); 


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the near
         * camera clipping plane.
         */
        void get_image_pt_on_near_clipping_plane
        (
            jwsc::Vector_f**      pt_out, 
            const jwsc::Vector_f* pt_in
        )
        throw (jwscxx::base::Arg_error); 


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the far
         * camera clipping plane.
         */
        void get_image_pt_on_far_clipping_plane
        (
            jwsc::Vector_f** pt_out, 
            float            x, 
            float            y
        ); 


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the far
         * camera clipping plane.
         */
        void get_image_pt_on_far_clipping_plane
        (
            jwsc::Vector_f**      pt_out, 
            const jwsc::Vector_f* pt_in
        )
        throw (jwscxx::base::Arg_error); 


        /** 
         * @brief Returns the aspect ration of width to height for the GL 
         * viewport.
         */
        static float get_gl_viewport_aspect();


        /** @brief Returns the viewport width in the GL. */
        static float get_gl_viewport_width();


        /** @brief Returns the viewport height in the GL. */
        static float get_gl_viewport_height();


        /** @brief Captures an image from the GL view (image) plane. */
        static void capture_gl_view(jwsc::Image_f** img_out);


        /** @brief Captures an image from the GL view (image) plane. */
        static void capture_gl_view(const char* fname) 
            throw (jwscxx::base::IO_error);


        /** @brief Captures an image from the GL view (image) plane. */
        static void capture_gl_view(const char* fname_base, uint32_t N) 
            throw (jwscxx::base::IO_error);


        /** @brief Reads this camera from an input stream. */
        virtual void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);


        /** @brief Writes this camera to an output stream. */
        virtual void write(std::ostream& out) const
            throw (jwscxx::base::IO_error);


    protected:

        /** @brief View reference point (center of the view-plane). */
        jwsc::Vector_f* vrp;

        /** @brief Projection reference point (center of the camera). */
        jwsc::Vector_f* prp;

        /** @brief View-right vector; x-axis of the view coordinates. */
        jwsc::Vector_f* u;

        /** @brief View-up vector; y-axis of the view coordinates. */
        jwsc::Vector_f* v;

        /** @brief View-plane normal; z-axis of the view coordinates. */
        jwsc::Vector_f* n;

        /** @brief Vertical aperture in degrees. */
        float aperture;

        /** @brief Focal length along the view-plane normal from the vrp. */
        float focal_length;

        /** @brief Half of the height of the view plane at the focal length. */
        float half_focal_height;

        /** @brief Half of the height of the near clipping plane. */
        float half_near_height;

        /** @brief Near clipping plane in the GL frustum. */
        float near;

        /** @brief Far clipping plane in the GL frustum. */
        float far;

        /** @brief Rotation matrix around the z-axis with angle phi. */
        jwsc::Matrix_f* R_phi;

        /** @brief Rotation matrix around the x-axis with angle theta. */
        jwsc::Matrix_f* R_theta;

        /** @brief Rotation matrix around the z-axis with angle psi. */
        jwsc::Matrix_f* R_psi;

        /** @brief Combined psi * theta * phi * theta^T * psi^T matrix. */
        jwsc::Matrix_f* R;
};




/**
 * @class Camera_d
 *
 * @brief Camera for 3D viewing in OpenGL.
 *
 * Uses double precision, floating point values.
 */
class Camera_d : public jwscxx::base::Cloneable, 
                 public jwscxx::base::Writeable, 
                 public jwscxx::base::Readable
{
    public:

        /** @brief Constructs a Camera_d. */
        Camera_d
        (
            double focal_pt_x,
            double focal_pt_y,
            double focal_pt_z,
            double ref_pt_x,
            double ref_pt_y,
            double ref_pt_z,
            double up_x,
            double up_y,
            double up_z,
            double aperture, 
            double clip_near, 
            double clip_far
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a Camera_d. */
        Camera_d
        (
            const jwsc::Vector_d* focal_pt,
            const jwsc::Vector_d* ref_pt,
            const jwsc::Vector_d* up,
            double                aperture, 
            double                clip_near, 
            double                clip_far
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a camera by copying another. */
        Camera_d(const Camera_d& c);


        /** @brief Reads a camera from an input file. */
        Camera_d(const char* fname) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Reads a camera from an input stream. */
        Camera_d(std::istream& in) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Deletes this Camera_d. */
        virtual ~Camera_d();


        /** @brief Copies a camera into this one. */
        virtual Camera_d& operator= (const Camera_d& c);


        /** @brief Clones this camera. */
        virtual Camera_d* clone() const;


        /** @brief Returns a constant pointer to the camera right (x) vector. */
        const jwsc::Vector_d* get_u() const;


        /** @brief Returns a constant pointer to the camera up (y) vector. */
        const jwsc::Vector_d* get_v() const;


        /** 
         * @brief Returns a constant pointer to the camera -view (z) vector.
         *
         * @note This vector is in the opposite direction as the view.
         */
        const jwsc::Vector_d* get_n() const;


        /** @brief Returns a constant pointer to the camera focal point. */
        const jwsc::Vector_d* get_prp() const;


        /** @brief Returns a constant pointer to the camera reference point. */
        const jwsc::Vector_d* get_vrp() const;


        /** @brief Returns the distance to the near clipping plane. */
        double get_clipping_near() const;


        /** @brief Returns the distance to the far clipping plane. */
        double get_clipping_far() const;


        /** @brief Returns the focal length. */
        double get_focal_length() const;


        /** @brief Returns the aperature. */
        double get_aperture() const;


        /** 
         * @brief Returns half the height of the image plane at the focal
         * length.
         */
        double get_half_focal_height() const;


        /** 
         * @brief Sets the focal length and updates the aperture by changing 
         * the left and right prp.
         */
        virtual void set_f_with_new_aperture_using_prp(double f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length using the same aperture by changing the 
         * left and right prp.
         */
        virtual void set_f_with_same_aperture_using_prp(double f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length and updates the aperture by changing 
         * the vrp.
         */
        virtual void set_f_with_new_aperture_using_vrp(double f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length using the same aperture by changing
         * the vrp.
         */
        virtual void set_f_with_same_aperture_using_vrp(double f) 
            throw (jwscxx::base::Arg_error);


        /** @brief Sets the vertical aperture. */
        virtual void set_aperture(double aperture) 
            throw (jwscxx::base::Arg_error);


        /** @brief Sets the clipping planes. */
        virtual void set_clipping(double clip_near, double clip_far) 
            throw (jwscxx::base::Arg_error);


        /** @brief Rotates this Camera_d around a vector. */
        virtual void rotate(double angle, double x, double y, double z) 
            throw (jwscxx::base::Arg_error);


        /** @brief Rotates this Camera_d around a vector. */
        virtual void rotate(double angle, const jwsc::Vector_d* r) 
            throw (jwscxx::base::Arg_error);


        /** @brief Translates this Camera_d. */
        virtual void translate(double x, double y, double z);


        /** @brief Translates this Camera_d. */
        virtual void translate(const jwsc::Vector_d* t) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the position, direction and focal length of the 
         * camera.
         */
        virtual void look_at
        (
            double focal_pt_x,
            double focal_pt_y,
            double focal_pt_z,
            double ref_pt_x,
            double ref_pt_y,
            double ref_pt_z,
            double up_x,
            double up_y,
            double up_z
        )
        throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the position, direction and focal length of the 
         * camera.
         */
        void look_at
        (
            const jwsc::Vector_d* focal_pt,
            const jwsc::Vector_d* ref_pt,
            const jwsc::Vector_d* up
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Sets the GL model view matrix with this Camera_d. */
        virtual void set_gl_modelview() const;


        /** @brief Sets the GL projection matrix with this Camera_d. */
        virtual void set_gl_projection() const;


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the camera
         * focal plane.
         */
        void get_image_pt_on_focal_plane
        (
            jwsc::Vector_d** pt_out, 
            double     x, 
            double     y
        ); 


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the camera
         * focal plane.
         */
        void get_image_pt_on_focal_plane
        (
            jwsc::Vector_d**      pt_out, 
            const jwsc::Vector_d* pt_in
        )
        throw (jwscxx::base::Arg_error); 


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the near
         * camera clipping plane.
         */
        void get_image_pt_on_near_clipping_plane
        (
            jwsc::Vector_d** pt_out, 
            double           x, 
            double           y
        ); 


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the near
         * camera clipping plane.
         */
        void get_image_pt_on_near_clipping_plane
        (
            jwsc::Vector_d**      pt_out, 
            const jwsc::Vector_d* pt_in
        )
        throw (jwscxx::base::Arg_error); 


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the far
         * camera clipping plane.
         */
        void get_image_pt_on_far_clipping_plane
        (
            jwsc::Vector_d** pt_out, 
            double           x, 
            double           y
        ); 


        /** 
         * @brief Calculates the R^3 world coordinates of a point on the far
         * camera clipping plane.
         */
        void get_image_pt_on_far_clipping_plane
        (
            jwsc::Vector_d**      pt_out, 
            const jwsc::Vector_d* pt_in
        )
        throw (jwscxx::base::Arg_error); 


        /** 
         * @brief Returns the aspect ration of width to height for the GL 
         * viewport.
         */
        static double get_gl_viewport_aspect();


        /** @brief Returns the viewport width in the GL. */
        static double get_gl_viewport_width();


        /** @brief Returns the viewport height in the GL. */
        static double get_gl_viewport_height();


        /** @brief Captures an image from the GL view (image) plane. */
        static void capture_gl_view(jwsc::Image_f** img_out);


        /** @brief Captures an image from the GL view (image) plane. */
        static void capture_gl_view(const char* fname) 
            throw (jwscxx::base::IO_error);


        /** @brief Captures an image from the GL view (image) plane. */
        static void capture_gl_view(const char* fname_base, uint32_t N) 
            throw (jwscxx::base::IO_error);


        /** @brief Reads this camera from an input stream. */
        virtual void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);


        /** @brief Writes this camera to an output stream. */
        virtual void write(std::ostream& out) const
            throw (jwscxx::base::IO_error);


    protected:

        /** @brief View reference point (center of the view-plane). */
        jwsc::Vector_d* vrp;

        /** @brief Projection reference point (center of the camera). */
        jwsc::Vector_d* prp;

        /** @brief View-right vector; x-axis of the view coordinates. */
        jwsc::Vector_d* u;

        /** @brief View-up vector; y-axis of the view coordinates. */
        jwsc::Vector_d* v;

        /** @brief View-plane normal; z-axis of the view coordinates. */
        jwsc::Vector_d* n;

        /** @brief Vertical aperture in degrees. */
        double aperture;

        /** @brief Focal length along the view-plane normal from the vrp. */
        double focal_length;

        /** @brief Half of the height of the view plane at the focal length. */
        double half_focal_height;

        /** @brief Half of the height of the near clipping plane. */
        double half_near_height;

        /** @brief Near clipping plane in the GL frustum. */
        double near;

        /** @brief Far clipping plane in the GL frustum. */
        double far;

        /** @brief Rotation matrix around the z-axis with angle phi. */
        jwsc::Matrix_d* R_phi;

        /** @brief Rotation matrix around the x-axis with angle theta. */
        jwsc::Matrix_d* R_theta;

        /** @brief Rotation matrix around the z-axis with angle psi. */
        jwsc::Matrix_d* R_psi;

        /** @brief Combined psi * theta * phi * theta^T * psi^T matrix. */
        jwsc::Matrix_d* R;
};




/**
 * @class Stereo_camera_f
 *
 * @brief Stereo pair of cameras for 3D viewing in OpenGL.
 *
 * Uses single precision, floating point values.
 */
class Stereo_camera_f : public Camera_f
{
    public:

        /** @brief Constructs a Stereo_camera_f. */
        Stereo_camera_f
        (
            float focal_pt_x,
            float focal_pt_y,
            float focal_pt_z,
            float ref_pt_x,
            float ref_pt_y,
            float ref_pt_z,
            float up_x,
            float up_y,
            float up_z,
            float aperture, 
            float clip_near, 
            float clip_far
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a Stereo_camera_f. */
        Stereo_camera_f
        (
            const jwsc::Vector_f* focal_pt,
            const jwsc::Vector_f* ref_pt,
            const jwsc::Vector_f* up,
            float                 aperture, 
            float                 clip_near, 
            float                 clip_far
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a camera by copying another. */
        Stereo_camera_f(const Stereo_camera_f& c);


        /** @brief Reads a camera from an input file. */
        Stereo_camera_f(const char* fname) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Reads a camera from an input stream. */
        Stereo_camera_f(std::istream& in) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Deletes this Camera_f. */
        virtual ~Stereo_camera_f();


        /** @brief Copies a camera into this one. */
        virtual Stereo_camera_f& operator= (const Stereo_camera_f& c);


        /** @brief Clones this camera. */
        virtual Stereo_camera_f* clone() const;


        /** @brief Returns a constant pointer to the left camera focal point. */
        const jwsc::Vector_f* get_left_prp() const;


        /** @brief Returns a constant pointer to the right camera focal point.*/
        const jwsc::Vector_f* get_right_prp() const;


        /** 
         * @brief Returns a constant pointer to the left camera reference 
         * point.
         */
        const jwsc::Vector_f* get_left_vrp() const;


        /** 
         * @brief Returns a constant pointer to the right camera reference 
         * point.
         */
        const jwsc::Vector_f* get_right_vrp() const;


        /** @brief Sets the vertical aperture. */
        void set_aperture(float aperture) throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length and updates the aperture by changing 
         * the left and right prp.
         */
        void set_f_with_new_aperture_using_prp(float f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length using the same aperture by changing the 
         * left and right prp.
         */
        void set_f_with_same_aperture_using_prp(float f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length and updates the aperture by changing 
         * the vrp.
         */
        void set_f_with_new_aperture_using_vrp(float f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length using the same aperture by changing
         * the vrp.
         */
        void set_f_with_same_aperture_using_vrp(float f) 
            throw (jwscxx::base::Arg_error);



        /** @brief Sets the clipping planes. */
        void set_clipping(float clip_near, float clip_far) 
            throw (jwscxx::base::Arg_error);


        /** @brief Rotates this Stereo_camera_f around a vector. */
        void rotate(float angle, float x, float y, float z) 
            throw (jwscxx::base::Arg_error);


        /** @brief Rotates this Stereo_camera_f around a vector. */
        void rotate(float angle, const jwsc::Vector_f* r) 
            throw (jwscxx::base::Arg_error);


        /** @brief Translates this Stereo_camera_f. */
        void translate(float x, float y, float z);


        /** @brief Translates this Stereo_camera_f. */
        void translate(const jwsc::Vector_f* t) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the position, direction and focal length of the 
         * camera.
         */
        void look_at
        (
            float focal_pt_x,
            float focal_pt_y,
            float focal_pt_z,
            float ref_pt_x,
            float ref_pt_y,
            float ref_pt_z,
            float up_x,
            float up_y,
            float up_z
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Sets the left GL model view matrix. */
        void set_left_gl_modelview();


        /** @brief Sets the right GL model view matrix. */
        void set_right_gl_modelview();


        /** @brief Sets the left GL projection matrix. */
        void set_left_gl_projection();


        /** @brief Sets the right GL projection matrix. */
        void set_right_gl_projection();


        /** @brief Reads this stereo camera from an input stream. */
        void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);


        /** @brief Writes this stereo camera to an output stream. */
        void write(std::ostream& out) const throw (jwscxx::base::IO_error);


    protected:

        /** @brief Left view reference point. */
        jwsc::Vector_f* left_vrp;

        /** @brief Right view reference point. */
        jwsc::Vector_f* right_vrp;

        /** @brief Left projection reference point. */
        jwsc::Vector_f* left_prp;

        /** @brief Right projection reference point. */
        jwsc::Vector_f* right_prp;

        /** @brief Seperation between the stereo cameras. */
        float separation;

        /** 
         * @brief Distance to the near clipping plane divided by the focal
         * length.
         */
        float ndfl;
};




/**
 * @class Stereo_camera_d
 *
 * @brief Stereo pair of cameras for 3D viewing in OpenGL.
 *
 * Uses single precision, floating point values.
 */
class Stereo_camera_d : public Camera_d
{
    public:

        /** @brief Constructs a Stereo_camera_d. */
        Stereo_camera_d
        (
            double focal_pt_x,
            double focal_pt_y,
            double focal_pt_z,
            double ref_pt_x,
            double ref_pt_y,
            double ref_pt_z,
            double up_x,
            double up_y,
            double up_z,
            double aperture, 
            double clip_near, 
            double clip_far
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a Stereo_camera_d. */
        Stereo_camera_d
        (
            const jwsc::Vector_d* focal_pt,
            const jwsc::Vector_d* ref_pt,
            const jwsc::Vector_d* up,
            double                aperture, 
            double                clip_near, 
            double                clip_far
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a camera by copying another. */
        Stereo_camera_d(const Stereo_camera_d& c);


        /** @brief Reads a camera from an input file. */
        Stereo_camera_d(const char* fname) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Reads a camera from an input stream. */
        Stereo_camera_d(std::istream& in) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Deletes this Camera_d. */
        virtual ~Stereo_camera_d();


        /** @brief Copies a camera into this one. */
        virtual Stereo_camera_d& operator= (const Stereo_camera_d& c);


        /** @brief Clones this camera. */
        virtual Stereo_camera_d* clone() const;


        /** @brief Returns a constant pointer to the left camera focal point. */
        const jwsc::Vector_d* get_left_prp() const;


        /** @brief Returns a constant pointer to the right camera focal point.*/
        const jwsc::Vector_d* get_right_prp() const;


        /** 
         * @brief Returns a constant pointer to the left camera reference 
         * point.
         */
        const jwsc::Vector_d* get_left_vrp() const;


        /** 
         * @brief Returns a constant pointer to the right camera reference 
         * point.
         */
        const jwsc::Vector_d* get_right_vrp() const;


        /** @brief Sets the vertical aperture. */
        void set_aperture(double aperture) throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length and updates the aperture by changing 
         * the left and right prp.
         */
        void set_f_with_new_aperture_using_prp(double f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length using the same aperture by changing the 
         * left and right prp.
         */
        void set_f_with_same_aperture_using_prp(double f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length and updates the aperture by changing 
         * the vrp.
         */
        void set_f_with_new_aperture_using_vrp(double f) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the focal length using the same aperture by changing
         * the vrp.
         */
        void set_f_with_same_aperture_using_vrp(double f) 
            throw (jwscxx::base::Arg_error);



        /** @brief Sets the clipping planes. */
        void set_clipping(double clip_near, double clip_far) 
            throw (jwscxx::base::Arg_error);


        /** @brief Rotates this Stereo_camera_d around a vector. */
        void rotate(double angle, double x, double y, double z) 
            throw (jwscxx::base::Arg_error);


        /** @brief Rotates this Stereo_camera_d around a vector. */
        void rotate(double angle, const jwsc::Vector_d* r) 
            throw (jwscxx::base::Arg_error);


        /** @brief Translates this Stereo_camera_d. */
        void translate(double x, double y, double z);


        /** @brief Translates this Stereo_camera_d. */
        void translate(const jwsc::Vector_d* t) 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Sets the position, direction and focal length of the 
         * camera.
         */
        void look_at
        (
            double focal_pt_x,
            double focal_pt_y,
            double focal_pt_z,
            double ref_pt_x,
            double ref_pt_y,
            double ref_pt_z,
            double up_x,
            double up_y,
            double up_z
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Sets the left GL model view matrix. */
        void set_left_gl_modelview();


        /** @brief Sets the right GL model view matrix. */
        void set_right_gl_modelview();


        /** @brief Sets the left GL projection matrix. */
        void set_left_gl_projection();


        /** @brief Sets the right GL projection matrix. */
        void set_right_gl_projection();


        /** @brief Reads this stereo camera from an input stream. */
        void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);


        /** @brief Writes this stereo camera to an output stream. */
        void write(std::ostream& out) const throw (jwscxx::base::IO_error);


    protected:

        /** @brief Left view reference point. */
        jwsc::Vector_d* left_vrp;

        /** @brief Right view reference point. */
        jwsc::Vector_d* right_vrp;

        /** @brief Left projection reference point. */
        jwsc::Vector_d* left_prp;

        /** @brief Right projection reference point. */
        jwsc::Vector_d* right_prp;

        /** @brief Seperation between the stereo cameras. */
        double separation;

        /** 
         * @brief Distance to the near clipping plane divided by the focal
         * length.
         */
        double ndfl;
};


}}


#endif
