/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Cuboid defined by a width, height and length.
 */


#ifndef CUBOID_H
#define CUBOID_H


#include <jwsc++/config.h>

#include <inttypes.h>

#include <jwsc/vector/vector.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/parapiped.h>


namespace jwscxx {
namespace graphics {


/** 
 * @class Cuboid_f
 *
 * @brief Cuboid defined by a width, height and length at the origin or a
 * specified postion.
 */
class Cuboid_f : public Parapiped_f
{
    public:

        /** @brief Constructs a cuboid at the origin. */
        Cuboid_f(float width, float height, float length)
            throw (jwscxx::base::Arg_error);


        /** @brief Constructs a cuboid at a particular location. */
        Cuboid_f
        (
            float width,  
            float height, 
            float length, 
            float x, 
            float y, 
            float z
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a cuboid at a particular location. */
        Cuboid_f
        (
            float width,  
            float height, 
            float length, 
            const jwsc::Vector_f* p
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a cuboid by copying another. */
        Cuboid_f(const Cuboid_f& c);


        /** @brief Reads a cuboid from an input file. */
        Cuboid_f(const char* fname) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Reads a cuboid from an input stream. */
        Cuboid_f(std::istream& in) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Deletes this cuboid. */
        virtual ~Cuboid_f();


        /** @brief Copies a cuboid into this one. */
        virtual Cuboid_f& operator= (const Cuboid_f& c);


        /** @brief Clones this cuboid. */
        virtual Cuboid_f* clone() const;
};


/** 
 * @class Cuboid_d
 *
 * @brief Cuboid defined by a width, height and length at the origin or a
 * specified postion.
 */
class Cuboid_d : public Parapiped_d
{
    public:

        /** @brief Constructs a cuboid at the origin. */
        Cuboid_d(double width, double height, double length)
            throw (jwscxx::base::Arg_error);


        /** @brief Constructs a cuboid at a particular location. */
        Cuboid_d
        (
            double width,  
            double height, 
            double length, 
            double x, 
            double y, 
            double z
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a cuboid at a particular location. */
        Cuboid_d
        (
            double width,  
            double height, 
            double length, 
            const jwsc::Vector_d* p
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a cuboid by copying another. */
        Cuboid_d(const Cuboid_d& c);


        /** @brief Reads a cuboid from an input file. */
        Cuboid_d(const char* fname) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Reads a cuboid from an input stream. */
        Cuboid_d(std::istream& in) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Deletes this cuboid. */
        virtual ~Cuboid_d();


        /** @brief Copies a cuboid into this one. */
        virtual Cuboid_d& operator= (const Cuboid_d& c);


        /** @brief Clones this cuboid. */
        virtual Cuboid_d* clone() const;
};


}}


#endif
