/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Robert Dionne
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Offscreen rendering buffer for X11 or Mac OS X.
 *
 * Under X11, pixmap and pbuffer are currently supported. For OS X, CGL pbuffers
 * are supported.
 */

#ifndef OFFSCREEN_H
#define OFFSCREEN_H


#include <jwsc++/config.h>

#include <jwsc++/base/exception.h>

#if defined JWSCXX_HAVE_X11 && defined JWSCXX_HAVE_GLX
#include <X11/Xlib.h>
#include <GL/gl.h>
#include <GL/glx.h>
#elif defined JWSCXX_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#include <OpenGL/gl.h>
#endif

#if defined JWSCXX_HAVE_OSMESA32
#include <GL/osmesa.h>
#endif


namespace jwscxx {
namespace graphics {


/**
 * @class Offscreen_buffer
 *
 * @brief Offscreen rendering buffer.
 */
class Offscreen_buffer
{
    public:

        /** @brief Constructs an offscreen buffer. */
        Offscreen_buffer(int width, int height) 
        { this->width = width; this->height = height; active = false; }

        /** @brief Deletes an offscreen buffer. */
        virtual ~Offscreen_buffer() { }

        /** @brief Activates an offscreen buffer. */
        virtual void activate() throw(jwscxx::base::Exception) = 0;

        /** @brief Deactivates an offscreen buffer. */
        virtual void deactivate() throw(jwscxx::base::Exception) = 0;

        /** @brief Reactivates an offscreen buffer. */
        virtual void reactivate() throw(jwscxx::base::Exception) = 0;

    protected:

        /** @brief The offscreen buffer is active. */
        bool active;

        /** @brief Width of the offscreen buffer. */
        int width; 

        /** @brief height of the offscreen buffer. */
        int height;
};


#if defined JWSCXX_HAVE_X11 && defined JWSCXX_HAVE_GLX
/**
 * @class GLX_offscreen_buffer
 *
 * @brief GLX Offscreen rendering buffer.
 */
class GLX_offscreen_buffer : public Offscreen_buffer
{
    public:

        /** @brief Constructs an offscreen buffer for GLX. */
        GLX_offscreen_buffer(int width, int height) 
            : Offscreen_buffer(width, height) { display_open = false; }

        /** @brief Deletes an offscreen buffer for GLX. */
        virtual ~GLX_offscreen_buffer() { }

        /** @brief Activates an offscreen buffer for GLX. */
        virtual void activate() throw(jwscxx::base::Exception);

        /** @brief Deactivates an offscreen buffer for GLX. */
        virtual void deactivate() throw(jwscxx::base::Exception);

        /** @brief Reactivates an offscreen buffer for GLX. */
        virtual void reactivate() throw(jwscxx::base::Exception);

    protected:

        /** 
         * @brief Saves the current GLX drawable object and context for later
         * restoration.
         */
        void save_old_context();

        /** @brief Opens the current X11 display. */
        void open_display(const char* disp_name=0) 
            throw(jwscxx::base::Result_error);

        /** @brief Creates the GLX drawable object and context. */
        virtual void create_drawable_and_context(int* attrs)
            throw(jwscxx::base::Exception) = 0;

        /** @brief Destroys the GLX drawable object. */
        virtual void destroy_drawable() = 0;

    protected:

        /** @brief The X11 display is open. */
        bool display_open;

        /** @brief The X11 display. */
        Display* display;

        /** @brief GLX drawable object. */
        GLXDrawable  drawable;
        
        /** @brief Previous GLX drawable object. */
        GLXDrawable old_drawable;

        /** @brief GLX context. */
        GLXContext context;

        /** @brief Previous GLX context. */
        GLXContext old_context;
};


/**
 * @class GLX_offscreen_pixmap
 *
 * @brief GLX offscreen rendering pixmap buffer.
 *
 * Uses the GLX pixmap extension for offscreen rendering.
 */
class GLX_offscreen_pixmap : public GLX_offscreen_buffer
{
    public:

        /** @brief Constructs an offscreen pixmap for GLX. */
        GLX_offscreen_pixmap(int width, int height, int* attrs=0)
            throw(jwscxx::base::Exception);
        
        /** @brief Deletes an offscreen pixmap for GLX. */
        virtual ~GLX_offscreen_pixmap();

    protected:

        /** @brief Creates the GLX drawable object and context. */
        virtual void create_drawable_and_context(int* attrs)
            throw(jwscxx::base::Result_error);

        /** @brief Destroys the GLX drawable object. */
        virtual void destroy_drawable();

    protected:

        /** @brief GLX pixmap to draw in. */
        Pixmap pixmap;
};


/**
 * @class GLX_offscreen_pbuffer
 *
 * @brief GLX offscreen rendering pbuffer buffer.
 *
 * Uses the SGI pbuffer extension for offscreen rendering.
 */
class GLX_offscreen_pbuffer : public GLX_offscreen_buffer
{
    public:

        /** @brief Constructs an offscreen pbuffer for GLX. */
        GLX_offscreen_pbuffer(int width, int height)
            throw(jwscxx::base::Exception);

        /** @brief Constructs an offscreen pbuffer for GLX. */
        GLX_offscreen_pbuffer(int width, int heigth, int* attrs)
            throw (jwscxx::base::Exception);

        /** @brief Deletes an offscreen pbuffer for GLX. */
        virtual ~GLX_offscreen_pbuffer();

    protected:

        /** @brief Checks if the X11 display has the pbuffer extension. */
        bool has_pbuffer_extension() const;

        /** @brief Creates the GLX drawable object and context. */
        virtual void create_drawable_and_context(int* attrs)
            throw(jwscxx::base::Exception);

        /** @brief Destroys the GLX drawable object. */
        virtual void destroy_drawable();
};


#elif defined JWSCXX_HAVE_OPENGL_FRAMEWORK
/**
 * @class CGL_offscreen_pbuffer
 *
 * @brief CGL Offscreen rendering buffer.
 */
class CGL_offscreen_pbuffer : public Offscreen_buffer
{
    public:

        /** @brief Constructs an offscreen pbuffer for CGL. */
        CGL_offscreen_pbuffer
        (
            GLsizei width, 
            GLsizei height, 
            GLenum  format=GL_RGBA,
            const CGLPixelFormatAttribute* attrs=0
        )
        throw (jwscxx::base::Exception);

        /** @brief Deletes an offscreen pbuffer for CGL. */
        virtual ~CGL_offscreen_pbuffer();

        /** @brief Activates an offscreen buffer for CGL. */
        virtual void activate() throw(jwscxx::base::Exception);

        /** @brief Deactivates an offscreen buffer for CGL. */
        virtual void deactivate() throw(jwscxx::base::Exception);

        /** @brief Reactivates an offscreen buffer for CGL. */
        virtual void reactivate() throw(jwscxx::base::Exception);

    protected:

        /** 
         * @brief Saves the current CGL context for later restoration.
         */
        virtual void save_old_context();

        /** @brief Creates the drawing context. */
        virtual void create_pbuffer_and_context
        (
            GLenum format,
            const CGLPixelFormatAttribute* attrs
        )
        throw(jwscxx::base::Exception);

    protected:

        /** @brief CGL context. */
        CGLContextObj context;

        /** @brief Previous CGL context. */
        CGLContextObj old_context;

        /** @brief CGL pbuffer. */
        CGLPBufferObj pbuffer;
};
#endif


#if defined JWSCXX_HAVE_OSMESA32
/**
 * @class OSMesa32_offscreen_buffer
 *
 * @brief OSMesa32 Offscreen rendering buffer.
 */
class OSMesa32_offscreen_buffer : public Offscreen_buffer
{
    public:

        /** @brief Constructs an offscreen buffer for OSMesa32. */
        OSMesa32_offscreen_buffer
        (
            GLsizei width, 
            GLsizei height, 
            GLenum  format=OSMESA_RGBA,
            GLint   depth_bits=4,
            GLint   stencil_bits=4,
            GLint   accum_bits=4
        )
        throw (jwscxx::base::Exception);

        /** @brief Deletes an offscreen buffer for OSMesa32. */
        virtual ~OSMesa32_offscreen_buffer();

        /** @brief Activates an offscreen buffer for OSMesa32. */
        virtual void activate() throw(jwscxx::base::Exception);

        /** @brief Deactivates an offscreen buffer for OSMesa32. */
        virtual void deactivate() throw(jwscxx::base::Exception);

        /** @brief Reactivates an offscreen buffer for OSMesa32. */
        virtual void reactivate() throw(jwscxx::base::Exception);

    protected:

        /** @brief Creates the drawing context. */
        virtual void create_buffer_and_context
        (
            GLenum format,
            GLint  depth_bits,
            GLint  stencil_bits,
            GLint  accum_bits
        )
        throw(jwscxx::base::Exception);

    protected:

        /** @brief OSMesa32 context. */
        OSMesaContext context;

        /** @brief OSMesa32 buffer. */
        void* buffer;
};
#endif


}}


#endif
