/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Parallelepiped: a hexahedron of which each face is a parallelegram.
 */


#ifndef PARAPIPED_H
#define PARAPIPED_H


#include <jwsc++/config.h>

#include <iostream>
#include <vector>

#include <inttypes.h>

#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>

#include <jwsc++/base/cloneable.h>
#include <jwsc++/base/readable.h>
#include <jwsc++/base/writeable.h>
#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/transformable.h>
#include <jwsc++/graphics/renderable.h>
#include <jwsc++/graphics/polymesh.h>
#include <jwsc++/graphics/polygon.h>


namespace jwscxx {
namespace graphics {


/** 
 * @class Parapiped_f
 *
 * @brief Parallelepiped: a hexahedron of which each face is a parallelegram.
 */
class Parapiped_f : public jwscxx::base::Cloneable, 
                    public jwscxx::base::Readable, 
                    public jwscxx::base::Writeable, 
                    public Transformable_f, 
                    public Wire_renderable, 
                    public Wire_occlude_renderable, 
                    public Solid_renderable,
                    public Polymesh_f
{
    public:

        /** @brief Constructs a parallelepiped. */
        Parapiped_f
        (
            float x1, float y1, float z1,
            float x2, float y2, float z2,
            float x3, float y3, float z3,
            float x4, float y4, float z4
        );


        /** @brief Constructs a parallelepiped. */
        Parapiped_f
        (
            const jwsc::Vector_f* p1, 
            const jwsc::Vector_f* p2, 
            const jwsc::Vector_f* p3, 
            const jwsc::Vector_f* p4
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a parallelepiped by copying another. */
        Parapiped_f(const Parapiped_f& p);


        /** @brief Reads a parallelepiped from an input file. */
        Parapiped_f(const char* fname) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Reads a parallelepiped from an input stream. */
        Parapiped_f(std::istream& in) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Deletes this parallelepiped. */
        virtual ~Parapiped_f();


        /** @brief Copies a parallelepiped into this one. */
        virtual Parapiped_f& operator= (const Parapiped_f& p);


        /** @brief Clones this parallelepiped. */
        virtual Parapiped_f* clone() const;


        /** @brief Returns the number of faces in this parallelepiped. */
        size_t num_faces() const;


        /** @brief Returns an indexed point defining this parapiped. */
        const jwsc::Vector_f* get_point(size_t i) const
            throw (jwscxx::base::Arg_error);


        /** @brief Returns the center vector for this parallelepiped. */
        const jwsc::Vector_f* get_center() const;


        /** @brief Returns an indexed parapiped face. */
        const jwscxx::graphics::Polygon_f* get_face(size_t i) const 
            throw (jwscxx::base::Arg_error);


        /** @brief Returns the faces of this parapiped. */
        const std::vector<const jwscxx::graphics::Polygon_f*>& get_faces() 
            const;


        /** @brief Sets the id of a face in the parapiped. */
        void set_face_id(size_t i, uint32_t id)
            throw (jwscxx::base::Arg_error);


        /** @brief Reads this parallelepiped from an input stream. */
        virtual void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);


        /** @brief Writes this parallelepiped to an output stream. */
        virtual void write(std::ostream& out) const
            throw (jwscxx::base::IO_error);


        /** @brief Applies a linear transformation to this parallelepiped. */
        virtual void transform(const jwsc::Matrix_f* M) 
            throw (jwscxx::base::Arg_error);

        
        /** @brief Renders this parallelepiped with the GL as a wire-frame. */
        virtual void wire_render() const;


        /** 
         * @brief Renders this parapiped with the GL as an occluded wire-frame 
         * into the depth buffer, to hide unseen lines.
         */
        virtual void wire_occlude_render() const;


        /** @brief Renders this parallelepiped with the GL as a solid. */
        virtual void solid_render() const;


        /** @brief Projects this parapiped into the current GL view. */
        virtual void project();


        /** @brief Returns the face adjacent to another on a shared edge. */
        uint32_t adjacent_face(uint32_t i, uint32_t p) const 
            throw (jwscxx::base::Arg_error);


    protected:

        /** @brief Polygons defining the parallelepiped. */
        std::vector<jwscxx::graphics::Polygon_f*> faces;

        /** @brief Points defining this parapiped. */
        std::vector<jwsc::Vector_f*> points;

        /** 
         * @brief Center of this parallelepiped.
         * 
         * Calculated as the average of the face centroids.
         */
        jwsc::Vector_f* center;

        /**
         * @brief Face adjacency matrix.
         *
         * Major index is the face, minor is the index of the first point (in
         * increasing order) on the face forming the adjacency edge.
         */
        jwsc::Matrix_u32* adjacency;
};


/** 
 * @class Parapiped_d
 *
 * @brief Parallelepiped: a hexahedron of which each face is a parallelegram.
 */
class Parapiped_d : public jwscxx::base::Cloneable, 
                    public jwscxx::base::Readable, 
                    public jwscxx::base::Writeable, 
                    public Transformable_d, 
                    public Wire_renderable, 
                    public Wire_occlude_renderable, 
                    public Solid_renderable,
                    public Polymesh_d
{
    public:

        /** @brief Constructs a parallelepiped. */
        Parapiped_d
        (
            double x1, double y1, double z1,
            double x2, double y2, double z2,
            double x3, double y3, double z3,
            double x4, double y4, double z4
        );


        /** @brief Constructs a parallelepiped. */
        Parapiped_d
        (
            const jwsc::Vector_d* p1, 
            const jwsc::Vector_d* p2, 
            const jwsc::Vector_d* p3, 
            const jwsc::Vector_d* p4
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Constructs a parallelepiped by copying another. */
        Parapiped_d(const Parapiped_d& p);


        /** @brief Reads a parallelepiped from an input file. */
        Parapiped_d(const char* fname) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Reads a parallelepiped from an input stream. */
        Parapiped_d(std::istream& in) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Deletes this parallelepiped. */
        virtual ~Parapiped_d();


        /** @brief Copies a parallelepiped into this one. */
        virtual Parapiped_d& operator= (const Parapiped_d& p);


        /** @brief Clones this parallelepiped. */
        virtual Parapiped_d* clone() const;


        /** @brief Returns the number of faces in this parallelepiped. */
        size_t num_faces() const;


        /** @brief Returns an indexed point defining this parapiped. */
        const jwsc::Vector_d* get_point(size_t i) const
            throw (jwscxx::base::Arg_error);


        /** @brief Returns the center vector for this parallelepiped. */
        const jwsc::Vector_d* get_center() const;


        /** @brief Returns an indexed parapiped face. */
        const jwscxx::graphics::Polygon_d* get_face(size_t i) const 
            throw (jwscxx::base::Arg_error);


        /** @brief Returns the faces of this parapiped. */
        const std::vector<const jwscxx::graphics::Polygon_d*>& get_faces() 
            const;


        /** @brief Sets the id of a face in the parapiped. */
        void set_face_id(size_t i, uint32_t id)
            throw (jwscxx::base::Arg_error);


        /** @brief Reads this parallelepiped from an input stream. */
        virtual void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);


        /** @brief Writes this parallelepiped to an output stream. */
        virtual void write(std::ostream& out) const
            throw (jwscxx::base::IO_error);


        /** @brief Applies a linear transformation to this parallelepiped. */
        virtual void transform(const jwsc::Matrix_d* M) 
            throw (jwscxx::base::Arg_error);

        
        /** @brief Renders this parallelepiped with the GL as a wire-frame. */
        virtual void wire_render() const;


        /** 
         * @brief Renders this parapiped with the GL as an occluded wire-frame 
         * into the depth buffer, to hide unseen lines.
         */
        virtual void wire_occlude_render() const;


        /** @brief Renders this parallelepiped with the GL as a solid. */
        virtual void solid_render() const;


        /** @brief Projects this parapiped into the current GL view. */
        virtual void project();


        /** @brief Returns the face adjacent to another on a shared edge. */
        uint32_t adjacent_face(uint32_t i, uint32_t p) const 
            throw (jwscxx::base::Arg_error);


    protected:

        /** @brief Polygons defining the parallelepiped. */
        std::vector<jwscxx::graphics::Polygon_d*> faces;

        /** @brief Points defining this parapiped. */
        std::vector<jwsc::Vector_d*> points;

        /** 
         * @brief Center of this parallelepiped.
         * 
         * Calculated as the average of the face centroids.
         */
        jwsc::Vector_d* center;

        /**
         * @brief Face adjacency matrix.
         *
         * Major index is the face, minor is the index of the first point (in
         * increasing order) on the face forming the adjacency edge.
         */
        jwsc::Matrix_u32* adjacency;
};


}}


#endif
