/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief 3D polygon with base position on the xy-plane centered at the
 * origin.
 */


#ifndef POLYGON_H
#define POLYGON_H


#include <jwsc++/config.h>

#include <iostream>
#include <vector>

#include <inttypes.h>

#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>

#include "jwsc++/base/cloneable.h"
#include "jwsc++/base/readable.h"
#include "jwsc++/base/writeable.h"
#include "jwsc++/base/exception.h"
#include "jwsc++/graphics/transformable.h"
#include "jwsc++/graphics/renderable.h"
#include "jwsc++/graphics/camera.h"


namespace jwscxx {
namespace graphics {


/** 
 * @class Polygon_f
 *
 * @brief 3D polygon in single precision floating point. 
 *
 * Base configuration of the polygon is centered on the origin of the 
 * xy-plane with normal vector aligned on the +z-axis.
 */
class Polygon_f : public jwscxx::base::Cloneable, 
                  public jwscxx::base::Readable, 
                  public jwscxx::base::Writeable, 
                  public Transformable_f, 
                  public Wire_renderable, 
                  public Wire_occlude_renderable,
                  public Solid_renderable
{
    public:

        /** @brief Constructs a polygon. */
        Polygon_f();


        /** @brief Constructs a polygon. */
        Polygon_f(uint32_t N);


        /** @brief Constructs a polygon by copying another. */
        Polygon_f(const Polygon_f& p);


        /** @brief Reads a polygon from an input file. */
        Polygon_f(const char* fname) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Reads a polygon from an input stream. */
        Polygon_f(std::istream& in) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Deletes this polygon. */
        virtual ~Polygon_f();


        /** @brief Copies a polygon into this polygon. */
        virtual Polygon_f& operator= (const Polygon_f& p);


        /** @brief Clones this polygon. */
        virtual Polygon_f* clone() const;


        /** @brief Reads this polygon from an input stream. */
        virtual void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);


        /** @brief Writes this polygon to an output stream. */
        virtual void write(std::ostream& out) const
            throw (jwscxx::base::IO_error);


        /** @brief Applies a linear transformation to this polygon. */
        virtual void transform(const jwsc::Matrix_f* M) 
            throw (jwscxx::base::Arg_error);

        
        /** @brief Renders this polygon with the GL as a wire-frame. */
        virtual void wire_render() const;


        /** 
         * @brief Renders this polygon with the GL as an occluded wire-frame 
         * into the depth buffer, to hide unseen lines.
         */
        virtual void wire_occlude_render() const;

        
        /** 
         * @brief Renders this polygon with the GL as a part of the silhouette,
         * into the depth buffer, to hide unseen lines.
         * The only edges that need to be rendered are those that are
         * atually part of the silhouette. iVector contains a boolean for
         * each edge, specifying whether that edge is part of the silhouette or not.
         * iBool is set to true when the face is visible and needs to be taken
         * into account for occlusions
         */
        virtual void wire_occluded_silhouette_render(std::vector<bool> & iVector, bool & iBool) const;


        /** @brief Renders this polygon with the GL as a solid. */
        virtual void solid_render() const;


        /** 
         * @brief Adds a point to the end of the collection of this polygon's
         * points.
         */
        void add_point(jwsc::Vector_f* pt) throw (jwscxx::base::Arg_error);


        /** 
         * @brief Adds a point to the end of the collection of this polygon's
         * points.
         */
        void add_point(float x, float y, float z) 
            throw (jwscxx::base::Arg_error);


        /** @brief Returns the normal vector for this polygon. */
        const jwsc::Vector_f* get_normal() const;


        /** @brief Returns the centroid vector for this polygon. */
        const jwsc::Vector_f* get_centroid() const;


        /** @brief Returns the number of points in this polygon. */
        size_t get_num_points() const;


        /** @brief Returns an indexed polygon point. */
        const jwsc::Vector_f* get_point(size_t i) const 
            throw (jwscxx::base::Arg_error);


        /** @brief Flips the direction of the normal vector. */
        void flip_normal();


        /** @brief Projects this polygon into the current GL view. */
        void project();


        /** @brief Returns whether this polygon is visible under a camera. */
        uint8_t is_visible(const Camera_f* camera) const;


        /** 
         * @brief Returns whether this polygon is visible under a camera.
         * This function returns true if the dot product between the
         * view vector and the surface normal is bigger than epsilon.
         * This is useful for determining polygons very close to the
         * silhouette that may occlude it.
         */
        uint8_t is_visible_up_to_epsilon(const Camera_d* camera, float & epsilon) const;


        /** @brief Returns the ID of the polygon. */
        uint32_t get_id() const;


        /** @brief Sets the ID of the polygon. */
        void set_id(uint32_t id);


    protected:

        /** @brief Updates the normal vector of this polygon. */
        void update_normal();

        /** @brief Updates the centroid vector of this polygon. */
        void update_centroid();


    protected:

        /** @brief 3D points defining the polygon. */
        std::vector<jwsc::Vector_f*> pts;

        /** @brief Normal vector for orientation. */
        jwsc::Vector_f* normal;

        /** @brief Center of mass of the polygon (assuming uniform density). */
        jwsc::Vector_f* centroid;

        /** 
         * @brief Indicates whether or not to flip the direction of the
         * normal.  
         */
        uint8_t normal_flipped;

        /** @brief Unique ID. */
        uint32_t id;


    private:

        /** @brief Generates a unique ID for a polygon. */
        static uint32_t generate_id();

};


/** 
 * @class Polygon_d
 *
 * @brief 3D polygon in double precision floating point. 
 *
 * Base configuration of the polygon is centered on the origin of the 
 * xy-plane with normal vector aligned on the +z-axis.
 */
class Polygon_d : public jwscxx::base::Cloneable, 
                  public jwscxx::base::Readable, 
                  public jwscxx::base::Writeable, 
                  public Transformable_d, 
                  public Wire_renderable, 
                  public Wire_occlude_renderable,
                  public Solid_renderable
{
    public:

        /** @brief Constructs a polygon. */
        Polygon_d();


        /** @brief Constructs a polygon. */
        Polygon_d(uint32_t N);


        /** @brief Constructs a polygon by copying another. */
        Polygon_d(const Polygon_d& p);


        /** @brief Reads a polygon from an input file. */
        Polygon_d(const char* fname) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Reads a polygon from an input stream. */
        Polygon_d(std::istream& in) throw (jwscxx::base::Arg_error,
                jwscxx::base::IO_error);


        /** @brief Deletes this polygon. */
        virtual ~Polygon_d();


        /** @brief Copies a polygon into this polygon. */
        virtual Polygon_d& operator= (const Polygon_d& p);


        /** @brief Clones this polygon. */
        virtual Polygon_d* clone() const;


        /** @brief Reads this polygon from an input stream. */
        virtual void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);


        /** @brief Writes this polygon to an output stream. */
        virtual void write(std::ostream& out) const
            throw (jwscxx::base::IO_error);


        /** @brief Applies a linear transformation to this polygon. */
        virtual void transform(const jwsc::Matrix_d* M) 
            throw (jwscxx::base::Arg_error);

        
        /** @brief Renders this polygon with the GL as a wire-frame. */
        virtual void wire_render() const;


        /** 
         * @brief Renders this polygon with the GL as an occluded wire-frame 
         * into the depth buffer, to hide unseen lines.
         */
        virtual void wire_occlude_render() const;

        /** 
         * @brief Renders this polygon with the GL as a part of the silhouette,
         * into the depth buffer, to hide unseen lines.
         * The only edges that need to be rendered are those that are
         * atually part of the silhouette. iVector contains a boolean for
         * each edge, specifying whether that edge is part of the silhouette or not.
         * iBool is set to true when the face is visible and needs to be taken
         * into account for occlusions
         */
        virtual void wire_occluded_silhouette_render(std::vector<bool> & iVector, bool & iBool) const;


        /** @brief Renders this polygon with the GL as a solid. */
        virtual void solid_render() const;


        /** 
         * @brief Adds a point to the end of the collection of this polygon's
         * points.
         */
        void add_point(jwsc::Vector_d* pt) throw (jwscxx::base::Arg_error);


        /** 
         * @brief Adds a point to the end of the collection of this polygon's
         * points.
         */
        void add_point(double x, double y, double z) 
            throw (jwscxx::base::Arg_error);


        /** @brief Returns the normal vector for this polygon. */
        const jwsc::Vector_d* get_normal() const;


        /** @brief Returns the centroid vector for this polygon. */
        const jwsc::Vector_d* get_centroid() const;


        /** @brief Returns the number of points in this polygon. */
        size_t get_num_points() const;


        /** @brief Returns an indexed polygon point. */
        const jwsc::Vector_d* get_point(size_t i) const 
            throw (jwscxx::base::Arg_error);


        /** @brief Flips the direction of the normal vector. */
        void flip_normal();


        /** @brief Projects this polygon into the current GL view. */
        void project();


        /** @brief Returns whether this polygon is visible under a camera. */
        uint8_t is_visible(const Camera_d* camera) const;

     
        /** 
         * @brief Returns whether this polygon is visible under a camera.
         * This function returns true if the dot product between the
         * view vector and the surface normal is bigger than epsilon.
         * This is useful for determining polygons very close to the
         * silhouette that may occlude it.
         */
        uint8_t is_visible_up_to_epsilon(const Camera_d* camera, double & epsilon) const;


        /** @brief Returns the ID of the polygon. */
        uint32_t get_id() const;


        /** @brief Sets the ID of the polygon. */
        void set_id(uint32_t id);


    protected:

        /** @brief Updates the normal vector of this polygon. */
        void update_normal();

        /** @brief Updates the centroid vector of this polygon. */
        void update_centroid();


    protected:

        /** @brief 3D points defining the polygon. */
        std::vector<jwsc::Vector_d*> pts;

        /** @brief Normal vector for orientation. */
        jwsc::Vector_d* normal;

        /** @brief Center of mass of the polygon (assuming uniform density). */
        jwsc::Vector_d* centroid;

        /** 
         * @brief Indicates whether or not to flip the direction of the
         * normal.  
         */
        uint8_t normal_flipped;

        /** @brief Unique ID. */
        uint32_t id;


    private:

        /** @brief Generates a unique ID for a polygon. */
        static uint32_t generate_id();

};


}}


#endif
