/*
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is by including this header.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @cclicense
 *
 * @brief Abstract class of connected polygons (faces) forming a mesh.
 */


#ifndef POLYMESH_H
#define POLYMESH_H

#include <inttypes.h>

#include <jwsc++/config.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/polygon.h>


namespace jwscxx {
namespace graphics {


/**
 * @class Polymesh_f
 *
 * @brief Abstract class of connected polygons (faces) forming a mesh.
 */
class Polymesh_f
{
    public:

        /** @brief Deletes this Transformable_f. */
        virtual ~Polymesh_f() { }


        /** @brief Returns the number of faces in this mesh. */
        virtual size_t num_faces() const = 0;


        /** 
         * @brief Returns an indexed face.
         *
         * @param  i  Index of the face to get; ranges 0 to N-1. 
         *
         * @return The specified face.
         *
         * @throw  jwscxx::base::Arg_error  Invalid face index.
         */
        const jwscxx::graphics::Polygon_f* get_face(size_t i) const 
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Returns the face index adjacent to another on a shared edge.
         *
         * The face point index must be in the range [0,N-1], where N is the
         * number of points in the @em ith face. To specify the edge formed
         * by the 0th and (N-1)th point, set p = N-1.
         *
         * @param  i  Face index to get the adjacent face of.
         * @param  p  Index of the first point in the @em ith face of the 
         *            shared edge. If set to (N-1), the other point forming the
         *            edge will be the 0th; face points are in increasing
         *            sequential order and loop around.
         *
         * @return The index of the face adjacent to the @em ith one in this
         * mesh.
         *
         * @throw  jwscxx::base::Arg_error  If i or p are not valid indices.
         */
        virtual uint32_t adjacent_face(uint32_t i, uint32_t p) const
            throw (jwscxx::base::Arg_error) = 0;
};


/**
 * @class Polymesh_d
 *
 * @brief Abstract class of connected polygons (faces) forming a mesh.
 */
class Polymesh_d
{
    public:

        /** @brief Deletes this Transformable_f. */
        virtual ~Polymesh_d() { }


        /** @brief Returns the number of faces in this mesh. */
        virtual size_t num_faces() const = 0;


        /** 
         * @brief Returns an indexed face.
         *
         * @param  i  Index of the face to get; ranges 0 to N-1. 
         *
         * @return The specified face.
         *
         * @throw  jwscxx::base::Arg_error  Invalid face index.
         */
        virtual const jwscxx::graphics::Polygon_d* get_face(size_t i) const 
            throw (jwscxx::base::Arg_error) = 0;


        /** 
         * @brief Returns the face index adjacent to another on a shared edge.
         *
         * The face point index must be in the range [0,N-1], where N is the
         * number of points in the @em ith face. To specify the edge formed
         * by the 0th and (N-1)th point, set p = N-1.
         *
         * @param  i  Face index to get the adjacent face of.
         * @param  p  Index of the first point in the @em ith face of the 
         *            shared edge. If set to (N-1), the other point forming the
         *            edge will be the 0th; face points are in increasing
         *            sequential order and loop around.
         *
         * @return The index of the face adjacent to the @em ith one in this
         * mesh.
         *
         * @throw  jwscxx::base::Arg_error  If i or p are not valid indices.
         */
        virtual uint32_t adjacent_face(uint32_t i, uint32_t p) const
            throw (jwscxx::base::Arg_error) = 0;
};


}}


#endif
